import os
import re

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import get_files


class FilePatchIsUpstreamStatusInactiveUpstreamDetails(Rule):
    def __init__(self):
        super().__init__(id='oelint.file.inactiveupstreamdetails',
                         severity='info',
                         message='Patch \'{FILE}\' with Upstream-Status: Inactive-Upstream has to have a lastcommit and/or lastrelease appended in []')

    def _get_recipe(self, items, path):
        # Find matching SRC_URI assignment
        return [x for x in items if x.VarValue.find(os.path.basename(path)) != -1]

    def check(self, _file, stash):
        res = []
        patches = get_files(stash, _file, '*.patch')
        _items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')

        _valid_class = {
            'Inactive-Upstream': r'Inactive-Upstream\s+\[\s*(lastcommit.+|lastrelease.+)+\s*\]',
        }

        for i in patches:
            with open(i) as _input:
                _recipe_match = self._get_recipe(_items, i)
                if _recipe_match:
                    _recipe_match = _recipe_match[0]
                else:
                    continue  # pragma: no cover
                try:
                    for m in re.finditer(r'^Upstream-Status:\s*(?P<class>.*)', _input.read(), re.MULTILINE):
                        for k, v in _valid_class.items():
                            if m.group('class').strip().startswith(k) and not re.match(v, m.group('class')):
                                res += self.finding(_recipe_match.Origin,
                                                    _recipe_match.InFileLine)
                except UnicodeDecodeError:  # pragma: no cover
                    pass  # pragma: no cover
        return res



import os

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Include
from oelint_parser.helper_files import expand_term
from oelint_parser.helper_files import find_local_or_in_layer


class FileIncludeNotFound(Rule):
    def __init__(self):
        super().__init__(id='oelint.file.includenotfound',
                         severity='warning',
                         message='\'{FILE}\' was not found')

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file,
                                      classifier=Include.CLASSIFIER):
            if item.Statement == 'include':
                _path = expand_term(stash, _file, item.IncName)
                if not find_local_or_in_layer(_path, os.path.dirname(item.Origin)):
                    res += self.finding(item.Origin, item.InFileLine,
                                        self.Msg.replace('{FILE}', item.IncName))
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarDependsSingleLine(Rule):
    def __init__(self):
        super().__init__(id='oelint.jetm.vars.dependssingleline',
                         severity='warning',
                         message='Each [R]DEPENDS entry should be put into a single line')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='DEPENDS')
        items += stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue='RDEPENDS')

        for i in items:
            if len(i.get_items(versioned=True)) > 1:
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarInconSpaces(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.inconspaces',
                         severity='error',
                         message='<FOO>')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            app_operation = i.AppendOperation()
            _stripped = i.VarValueStripped.lstrip(chr(0x1b))
            # allow 'spaceless' append to FILESEXTRAPATHS as there is
            # no operation which supports the combination of
            # append + :=
            if 'append' in app_operation and not _stripped.startswith(' ') and i.VarName in ['FILESEXTRAPATHS']:
                continue
            if ' += ' in app_operation and i.VarValueStripped.startswith(' '):
                res += self.finding(i.Origin, i.InFileLine,
                                    'Assignment should be \'VAR += "foo"\' not \'VAR += " foo"\'')
            if 'append' in app_operation and not _stripped.startswith(' '):
                res += self.finding(i.Origin, i.InFileLine,
                                    'Assignment should be \'VAR_append = " foo"\' not \'VAR_append = "foo"\'')
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_scr_components
from oelint_parser.parser import INLINE_BLOCK


class VarSRCUriOptions(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.srcuridomains',
                         severity='warning',
                         message='Recipe is pulling from different domains, this will likely cause issues')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')

        _domains = set()
        for item in items:
            for u in [x.strip('\'').strip() for x in item.get_items()]:
                if u == INLINE_BLOCK:
                    continue
                _url = get_scr_components(u)
                if _url['scheme'] and _url['scheme'] not in ['file']:
                    _domains.add(_url['src'].split('/')[0])
        if len(_domains) > 1:
            res += self.finding(item.Origin, item.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import expand_term
from oelint_parser.helper_files import get_scr_components
from oelint_parser.helper_files import guess_recipe_version


class VarsDownloadfilename(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.downloadfilename',
                         severity='warning',
                         message='Fetcher does create a download artifact without \'PV\' in the filename')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')

        _pv = expand_term(stash, _file, '${PV}') or guess_recipe_version(_file)

        for i in items:
            for item in [expand_term(stash, _file, x) for x in i.get_items()]:
                _scrcomp = get_scr_components(item)
                if _scrcomp.get('scheme', '') in ['https', 'http', 'ftp']:
                    _src = _scrcomp.get('src', '')
                    _dfn = _scrcomp.get('options', {}).get(
                        'downloadfilename', '')

                    if _pv not in _src:
                        if _dfn and _pv in _dfn:
                            continue
                        res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarBugtrackerIsUrl(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.fileextrapathsop',
                         severity='error',
                         message='\'FILESEXTRAPATHS\' should only be used in combination with \' := \'')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        for i in items:
            if i.VarName in ['FILESEXTRAPATHS_prepend', 'FILESEXTRAPATHS_append', 'FILESEXTRAPATHS']:
                if i.VarOp != ' := ':
                    res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarSummaryLinebreaks(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.summarylinebreaks',
                         severity='warning',
                         message='\'SUMMARY\' should not contain line breaks')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='SUMMARY')
        for i in items:
            if i.Raw.strip().find('\\n') != -1:
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.constants import CONSTANTS


class VarQuoted(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.bbvars',
                         severity='warning',
                         message='Variable \'{VAR}\' should be set on a disto/layer or local.conf level, not in a recipe',
                         appendix=CONSTANTS.VariablesProtected)

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in [x for x in items if x.VarName in CONSTANTS.VariablesProtected]:
            if i.VarOp not in [' ??= ', ' ?= ']:
                res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.replace(
                    '{VAR}', i.VarName), appendix=i.VarName)
        return res



import importlib
import inspect
import os
import pkgutil
import sys

from colorama import Style

from oelint_adv.color import get_color_by_severity
from oelint_adv.color import get_colorize
from oelint_adv.rule_file import get_messageformat
from oelint_adv.rule_file import get_noinfo
from oelint_adv.rule_file import get_nowarn
from oelint_adv.rule_file import get_relpaths
from oelint_adv.rule_file import get_rulefile
from oelint_adv.rule_file import get_suppressions


class Rule:
    def __init__(self, id='', severity='', message='', onappend=True, onlyappend=False, appendix=()):
        """constructor

        Keyword Arguments:
            id {str} -- ID of the rule (default: {''})
            severity {str} -- severity of the rule (default: {''})
            message {str} -- Rule message (default: {''})
            onappend {bool} -- true if rule should be run on bbappends (default: {True})
            onlyappend {bool} -- true if rule applies to bbappends only (default: {False})
            appendix {tuple} -- possible appendix to id
        """
        self.ID = id
        self.Severity = severity
        self.Msg = message
        self.OnAppend = onappend
        self.OnlyAppend = onlyappend
        self.Appendix = appendix

    def check(self, _file, stash):
        """Stub for running check - is overridden by each rule

        Arguments:
            _file {str} -- File to be parsed
            stash {oelint_adv.cls_stash.Stash} -- Parsed stash

        Returns:
            list -- List of findings
        """
        return []  # pragma: no cover

    def fix(self, _file, stash):
        """Stub for fix function - can be overridden by each rule

        Arguments:
            _file {str} -- File to be parsed
            stash {oelint_adv.cls_stash.Stash} -- Parsed stash

        Returns:
            list -- list of changed files
        """
        return []

    def finding(self, _file, _line, override_msg=None, appendix=None):
        """Called by rule to indicate a finding

        Arguments:
            _file {str} -- Full path to file or origin
            _line {int} -- Line number in file

        Keyword Arguments:
            override_msg {str} -- Optional string which overrides the set standard message (default: {None})
            appendix {str} -- Optional appendix to rule ID (default: {None})

        Returns:
            str -- Human readable finding (possibly with color codes)
        """
        if not self.OnAppend and _file.endswith('.bbappend'):
            return []  # pragma: no cover
        if self.OnlyAppend and not _file.endswith('.bbappend'):
            return []  # pragma: no cover
        if override_msg is None:
            override_msg = self.Msg
        _id = [self.ID]
        _display_id = self.ID
        if appendix:
            _id.append(self.ID + '.' + appendix)
            _display_id += '.' + appendix
        # filter out suppressions
        if any(x in get_suppressions() for x in _id):
            return []
        _severity = self.get_severity(appendix)
        if _severity is None:
            # the rule is disabled
            return []
        if _severity == 'info' and get_noinfo():
            return []
        if _severity == 'warning' and get_nowarn():
            return []
        if _line <= 0:
            # Fix those issues, that don't come with a line
            _line = 1

        _path = os.path.abspath(_file)
        if get_relpaths():
            _path = os.path.relpath(_path, os.getcwd())

        _style = ''
        _color = ''
        if get_colorize():
            _color = get_color_by_severity(_severity)
            _style = Style.RESET_ALL

        _msg = get_messageformat().format(path=_path, line=_line, severity=_severity,
                                          id=_display_id, msg=override_msg)

        return [(_line, f'{_color}{_msg}{_style}')]

    def __repr__(self):
        return '{id}'.format(id=self.ID)  # pragma: no cover

    def get_severity(self, appendix=None):
        """Get the configured severity for this rule, if it is enabled.

        Keyword Arguments:
            appendix {str} -- Potential subrule name (default: {None})

        Returns:
            str -- Severity for this rule if it is enabled, {None} if disabled.
        """
        _rule_file = get_rulefile()
        if not _rule_file:
            return self.Severity
        _subid = None if appendix is None else f'{self.ID}.{appendix}'
        if _subid and _subid in _rule_file:
            _severity = _rule_file[_subid]
        elif self.ID in _rule_file:
            _severity = _rule_file[self.ID]
        else:
            # rule not in rulefile
            return None
        return _severity if _severity != '' else self.Severity

    def get_ids(self):
        """Returns all possible IDs of the rule

        Returns:
            list -- possible IDS of the rule
        """
        return [self.ID] + ['{id}.{app}'.format(id=self.ID, app=x) for x in self.Appendix]

    def get_rulefile_entries(self):
        """Returns a dictionary of entries which would represent the currently
        enabled ruleset (for this rule) in a rulefile.

        Returns:
            dict -- list of rulefile entries
        """
        return {
            **({} if self.get_severity() is None else {self.ID: self.get_severity()}),
            **{f'{self.ID}.{x}': self.get_severity(x) for x in self.Appendix if self.get_severity(x) is not None},
        }

    def format_message(self, *args, **kwargs):
        """Format message

        Returns:
            str -- formatted message
        """
        return self.Msg.format(*args, **kwargs)

    @staticmethod
    def is_lone_append(stash, file):
        """Check if the file is a bbappend file without any matching
           bb file

        Args:
            stash {oelint_parser.cls_stash.Stash} -- Parsed stash
            file {str} -- Path to file

        Returns:
            bool: True if bbappend has no matches in stash
        """
        if not file.endswith('.bbappend'):
            return False
        for item in stash.GetItemsFor(filename=file):
            if any(x.endswith('.bb') for x in item.Links):  # pragma: no cover
                return False  # pragma: no cover
        return True


def load_rules(args, add_rules=(), add_dirs=()):
    """Load rules from set directories

    Keyword Arguments:
        add_rules {tuple} -- Additional builtin rulesets to be loaded (default: {[]})
        add_dirs {tuple} -- Additional directories to parse for rules (default: {[]})

    Returns:
        list -- Class instances of loaded rules
    """
    res = []
    _rule_file = get_rulefile()
    _path_list = {
        'base': {'path': 'rule_base', 'builtin': True},
    }
    for ar in add_rules:
        _path_list[ar] = {'path': 'rule_{a}'.format(a=ar), 'builtin': True}
    for ar in add_dirs:
        _path_list['additional_{a}'.format(a=os.path.basename(ar))] = {
            'path': ar, 'builtin': False}  # pragma: no cover
    for _, v in _path_list.items():
        if v['builtin']:
            _searchpath = os.path.join(os.path.dirname(
                os.path.abspath(__file__)), v['path'])
        else:
            _searchpath = os.path.join(v['path'])  # pragma: no cover
            sys.path.append(os.path.dirname(v['path']))  # pragma: no cover
        packages = pkgutil.walk_packages(path=[_searchpath])
        for _, name, _ in packages:
            if v['builtin']:
                name = __name__.split('.')[0] + '.' + v['path'] + '.' + name
            else:
                name = os.path.basename(v['path']) + '.' + name  # pragma: no cover
            try:
                mod = importlib.import_module(name)
                for m in inspect.getmembers(mod, inspect.isclass):
                    try:
                        if issubclass(m[1], Rule):
                            inst = m[1]()
                            _potential_ids = [
                                inst.ID] + ['{a}.{b}'.format(a=inst.ID, b=x) for x in inst.Appendix]
                            if any(_potential_ids):
                                if _rule_file and not any(x in _rule_file for x in _potential_ids):
                                    continue  # pragma: no cover
                                res.append(inst)
                    except Exception:  # pragma: no cover
                        pass  # pragma: no cover
            except Exception as e:  # pragma: no cover
                if not args.quiet:  # pragma: no cover
                    print(  # noqa: T001 this print is fine here
                        'Can\'t load rule {rule} -> {exp}'.format(rule=name, exp=e))  # pragma: no cover
    return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import get_scr_components
from oelint_parser.parser import INLINE_BLOCK


class VarSRCUriGitTag(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.srcurifile',
                         severity='warning',
                         message='First item of SRC_URI should not be a file:// fetcher, if multiple fetcher are used')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')
        _fetcher = []
        for item in items:
            if any([item.Flag.endswith(x) for x in ['md5sum', 'sha256sum']]):
                # These are just the hashes
                continue
            lines = [y.strip('"') for y in item.get_items() if y]

            for x in lines:
                if x == INLINE_BLOCK:
                    _fetcher.append(('inline', item.InFileLine))
                    continue
                _url = get_scr_components(x)
                if _url['scheme']:
                    _fetcher.append((_url['scheme'], item.InFileLine))
        if _fetcher:
            if any(x[0] != 'file' for x in _fetcher) and _fetcher[0][0] == 'file':
                res += self.finding(item.Origin, _fetcher[0][1])
        return res



import argparse
import json
import os
import re
import sys
from configparser import ConfigParser
from configparser import NoOptionError
from configparser import NoSectionError
from configparser import ParsingError

from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import load_rules
from oelint_adv.color import set_colorize
from oelint_adv.rule_file import set_messageformat
from oelint_adv.rule_file import set_noinfo
from oelint_adv.rule_file import set_nowarn
from oelint_adv.rule_file import set_relpaths
from oelint_adv.rule_file import set_rulefile
from oelint_adv.rule_file import set_suppressions

sys.path.append(os.path.abspath(os.path.join(__file__, '..')))


class TypeSafeAppendAction(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest) or []
        if isinstance(items, str):
            items = re.split(r'\s+|\t+|\n+', items)  # pragma: no cover
        items.append(values)  # pragma: no cover
        setattr(namespace, self.dest, items)  # pragma: no cover


def parse_configfile():
    config = ConfigParser()
    for conffile in [os.environ.get('OELINT_CONFIG', '/does/not/exist'),
                     os.path.join(os.getcwd(), '.oelint.cfg'),
                     os.path.join(os.environ.get('HOME', '/does/not/exist'), '.oelint.cfg')]:
        try:
            if not os.path.exists(conffile):
                continue
            config.read(conffile)
            return {k.replace('-', '_'): v for k, v in config.items('oelint')}
        except (PermissionError, SystemError) as e:  # pragma: no cover
            print(f'Failed to load config file {conffile}. {e!r}')  # noqa: T001 - it's fine here; # pragma: no cover
        except (NoSectionError, NoOptionError, ParsingError) as e:
            print(f'Failed parsing config file {conffile}. {e!r}')  # noqa: T001 - it's here for a reason
    return {}


def create_argparser():
    parser = argparse.ArgumentParser(prog='oelint-adv',
                                     description='Advanced OELint - Check bitbake recipes against OECore styleguide')
    parser.register('action', 'tsappend', TypeSafeAppendAction)
    parser.add_argument('--suppress', default=[],
                        action='tsappend', help='Rules to suppress')
    parser.add_argument('--output', default=sys.stderr,
                        help='Where to flush the findings (default: stderr)')
    parser.add_argument('--fix', action='store_true', default=False,
                        help='Automatically try to fix the issues')
    parser.add_argument('--nobackup', action='store_true', default=False,
                        help='Don\'t create backup file when auto fixing')
    parser.add_argument('--addrules', nargs='+', default=[],
                        help='Additional non-default rulessets to add')
    parser.add_argument('--customrules', nargs='+', default=[],
                        help='Additional directories to parse for rulessets')
    parser.add_argument('--rulefile', default=None,
                        help='Rulefile')
    parser.add_argument('--constantfile', default=None, help='Constantfile')
    parser.add_argument('--color', action='store_true', default=False,
                        help='Add color to the output based on the severity')
    parser.add_argument('--quiet', action='store_true', default=False,
                        help='Print findings only')
    parser.add_argument('--noinfo', action='store_true', default=False,
                        help='Don\'t print information level findings')
    parser.add_argument('--nowarn', action='store_true', default=False,
                        help='Don\'t print warning level findings')
    parser.add_argument('--relpaths', action='store_true', default=False,
                        help='Show relative paths instead of absolute paths in results')
    parser.add_argument('--noid', action='store_true', default=False,
                        help='Don\'t show the error-ID in the output')
    parser.add_argument('--messageformat', default='{path}:{line}:{severity}:{id}:{msg}',
                        type=str, help='Format of message output')
    parser.add_argument('--constantmods', default=[], nargs='+',
                        help='''
                             Modifications to the constant db.
                             prefix with:
                             + - to add to DB,
                             - - to remove from DB,
                             None - to override DB
                            ''')
    parser.add_argument('--print-rulefile', action='store_true', default=False,
                        help='Print loaded rules as a rulefile and exit')
    parser.add_argument('--exit-zero', action='store_true', default=False,
                        help='Always return a 0 (non-error) status code, even if lint errors are found')
    # Override the defaults with the values from the config file
    parser.set_defaults(**parse_configfile())

    parser.add_argument('files', nargs='*', help='File to parse')

    return parser


def parse_arguments():
    return create_argparser().parse_args()  # pragma: no cover


def arguments_post(args):  # noqa: C901 - complexity is still okay
    # Convert boolean symbols
    for _option in [
        'color',
        'exit_zero',
        'fix',
        'nobackup',
        'noinfo',
        'nowarn',
        'print_rulefile',
        'quiet',
        'relpaths',
    ]:
        try:
            setattr(args, _option, bool(getattr(args, _option)))
        except AttributeError:  # pragma: no cover
            pass  # pragma: no cover

    # Convert list symbols
    for _option in [
        'suppress',
        'constantmods',
    ]:
        try:
            if not isinstance(getattr(args, _option), list):
                setattr(args, _option, [x.strip() for x in (getattr(args, _option) or '').split('\n') if x])
        except AttributeError:  # pragma: no cover
            pass  # pragma: no cover

    if args.files == [] and not args.print_rulefile:
        raise argparse.ArgumentTypeError('no input files')

    if args.rulefile:
        try:
            with open(args.rulefile) as i:
                set_rulefile(json.load(i))
        except (FileNotFoundError, json.JSONDecodeError):
            raise argparse.ArgumentTypeError(
                '\'rulefile\' is not a valid file')

    if args.constantfile:
        try:
            with open(args.constantfile) as i:
                CONSTANTS.AddFromConstantFile(json.load(i))
        except (FileNotFoundError, json.JSONDecodeError):
            raise argparse.ArgumentTypeError(
                '\'constantfile\' is not a valid file')

    for mod in args.constantmods:
        try:
            with open(mod.lstrip('+-')) as _in:
                _cnt = json.load(_in)
            if mod.startswith('+'):
                CONSTANTS.AddConstants(_cnt)
            elif mod.startswith('-'):
                CONSTANTS.RemoveConstants(_cnt)
            else:
                CONSTANTS.OverrideConstants(_cnt)
        except (FileNotFoundError, json.JSONDecodeError):
            raise argparse.ArgumentTypeError(
                'mod file \'{file}\' is not a valid file'.format(file=mod))

    set_colorize(args.color)
    set_nowarn(args.nowarn)
    set_noinfo(args.noinfo)
    set_relpaths(args.relpaths)
    set_suppressions(args.suppress)
    if args.noid:
        # just strip id from message format if noid is requested
        args.messageformat = args.messageformat.replace('{id}', '')
        # strip any double : resulting from the previous operation
        args.messageformat = args.messageformat.replace('::', ':')
    set_messageformat(args.messageformat)
    return args


def group_files(files):
    # in case multiple bb files are passed at once we might need to group them to
    # avoid having multiple, potentially wrong hits of include files shared across
    # the bb files in the stash
    res = {}
    for f in files:
        _filename, _ext = os.path.splitext(f)
        if _ext not in ['.bb']:
            continue
        if '_' in os.path.basename(_filename):
            _filename_key = '_'.join(os.path.basename(
                _filename).split('_')[:-1]).replace('%', '')
        else:
            _filename_key = os.path.basename(_filename)
        if _filename_key not in res:  # pragma: no cover
            res[_filename_key] = set()
        res[_filename_key].add(f)

    # second round now for the bbappend files
    for f in files:
        _filename, _ext = os.path.splitext(f)
        if _ext not in ['.bbappend']:
            continue
        _match = False
        for _, v in res.items():
            _needle = '.*/' + os.path.basename(_filename).replace('%', '.*')
            if any(re.match(_needle, x) for x in v):
                v.add(f)
                _match = True
                break
        if not _match:
            _filename_key = '_'.join(os.path.basename(
                _filename).split('_')[:-1]).replace('%', '')
            if _filename_key not in res:  # pragma: no cover
                res[_filename_key] = set()
            res[_filename_key].add(f)

    # as sets are unordered, we convert them to sorted lists at this point
    # order is like the files have been passed via CLI
    for k, v in res.items():
        res[k] = sorted(v, key=lambda index: files.index(index))

    return res.values()


def print_rulefile(args):
    rules = load_rules(args, add_rules=args.addrules,
                       add_dirs=args.customrules)
    ruleset = {}
    for r in rules:
        ruleset.update(r.get_rulefile_entries())
    print(json.dumps(ruleset, indent=2))  # noqa: T001 - it's here for a reason


def run(args):
    try:
        rules = load_rules(args, add_rules=args.addrules,
                           add_dirs=args.customrules)
        _loaded_ids = []
        for r in rules:
            _loaded_ids += r.get_ids()
        if not args.quiet:
            print('Loaded rules:\n\t{rules}'.format(  # noqa: T001 - it's here for a reason
                rules='\n\t'.join(sorted(_loaded_ids))))
        issues = []
        fixedfiles = []
        groups = group_files(args.files)
        for group in groups:
            stash = Stash(args)
            for f in group:
                try:
                    stash.AddFile(f)
                except FileNotFoundError as e:  # pragma: no cover
                    if not args.quiet:  # pragma: no cover
                        print('Can\'t open/read: {e}'.format(e=e))  # noqa: T001 - it's fine here; # pragma: no cover

            stash.Finalize()

            _files = list(set(stash.GetRecipes() + stash.GetLoneAppends()))
            for _, f in enumerate(_files):
                for r in rules:
                    if not r.OnAppend and f.endswith('.bbappend'):
                        continue
                    if r.OnlyAppend and not f.endswith('.bbappend'):
                        continue
                    if args.fix:
                        fixedfiles += r.fix(f, stash)
                    issues += r.check(f, stash)
            fixedfiles = list(set(fixedfiles))
            for f in fixedfiles:
                _items = [f] + stash.GetLinksForFile(f)
                for i in _items:
                    items = stash.GetItemsFor(filename=i, nolink=True)
                    if not args.nobackup:
                        os.rename(i, i + '.bak')  # pragma: no cover
                    with open(i, 'w') as o:
                        o.write(''.join([x.RealRaw for x in items]))
                        if not args.quiet:
                            print('{path}:{lvl}:{msg}'.format(path=os.path.abspath(i),  # noqa: T001 - it's fine here; # pragma: no cover
                                                    lvl='debug', msg='Applied automatic fixes'))

        return sorted(set(issues), key=lambda x: x[0])
    except Exception:
        import traceback

        # pragma: no cover
        print('OOPS - That shouldn\'t happen - {files}'.format(files=args.files))   # noqa: T001 - it's here for a reason
        # pragma: no cover
        traceback.print_exc()
    return []


def main():  # pragma: no cover
    args = arguments_post(parse_arguments())

    if args.print_rulefile:
        print_rulefile(args)
        sys.exit(0)

    issues = run(args)

    if args.output != sys.stderr:
        args.output = open(args.output, 'w')
    args.output.write('\n'.join([x[1] for x in issues]))
    if issues:
        args.output.write('\n')
    if args.output != sys.stderr:
        args.output.close()

    exit_code = len(issues) if not args.exit_zero else 0
    sys.exit(exit_code)


if __name__ == '__main__':
    main()  # pragma: no cover



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Include
from oelint_parser.cls_item import Variable


class VarDependsAppend(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.dependsappend',
                         severity='error',
                         message='DEPENDS should only be appended, not overwritten as an include or inherit')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='DEPENDS')
        incinh = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='inherit', nolink=True)
        incinh += stash.GetItemsFor(filename=_file, classifier=Include.CLASSIFIER, nolink=True)

        earliest = 99999999
        if incinh:
            earliest = min(x.Line for x in incinh)

        for i in items:
            if not i.AppendOperation() and i.Line > earliest:
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarQuoted(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.valuequoted',
                         severity='error',
                         message='Variable value should be quoted')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if i.VarName == 'inherit':
                continue
            # Don't use VarValueStripped here as we explicitly want the quotes
            # at the beginning and the end of the value
            val = i.VarValue.strip()
            if not val.startswith('"') or not val.endswith('"'):
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarPnUsageDiscouraged(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.pnusagediscouraged',
                         severity='warning',
                         message='Variable shouldn\'t contain ${PN} or ${BPN}')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        needles = ['SUMMARY', 'HOMEPAGE', 'BUGTRACKER', 'DESCRIPTION']
        for i in [x for x in items if x.VarName in needles]:
            if '${PN}' in i.VarValue or '${BPN}' in i.VarValue:
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarSpacesOnAssignment(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.spacesassignment',
                         severity='warning',
                         message='Suggest spaces around variable assignment. E.g. \'FOO = "BAR"\'')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if i.VarName == 'inherit':
                continue
            if i.VarOp not in Variable.VAR_VALID_OPERATOR:
                res += self.finding(i.Origin, i.InFileLine)
        return res



import ast

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Function


class TaskPythonPrefix(Rule):
    def __init__(self):
        super().__init__(id='oelint.task.pythonprefix',
                         severity='warning',
                         message='Tasks containing python code, should be prefixed with python in function header')

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER):
            # Don't make assumptions about functions that consist of single
            # line only
            if len([x for x in item.FuncBodyRaw.split('\n') if x.strip()]) < 2:
                continue
            try:
                ast.parse(item.FuncBodyRaw, 'tempfile')
                if not item.IsPython:
                    res += self.finding(item.Origin, item.InFileLine)
            except Exception:
                pass  # noqa: S110 - intentionally ignore all errors
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_layer_root


class FileNoSpaces(Rule):
    def __init__(self):
        super().__init__(id='oelint.file.nospaces',
                         severity='error',
                         message='Path to file contains spaces. Please remove them')

    def check(self, _file, stash):
        res = []
        _layer_root = get_layer_root(_file)
        _relpath = _file.replace(_layer_root, '').lstrip('/')
        if ' ' in _relpath:
            res += self.finding(_file, 1)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarListAppend(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.listappend',
                         severity='error',
                         message='<FOO>')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        needles = ['PACKAGES', 'SRC_URI', 'FILES', 'RDEPENDS', 'DEPENDS']

        for i in items:
            if not any(i.VarName.startswith(x) for x in needles):
                continue
            if i.VarName.startswith('FILESEXTRAPATHS'):
                # Caught by the `FILES` above but this list is colon separated.
                continue
            ops = i.AppendOperation()
            if not i.VarValue.startswith('" ') and any(x in ops for x in ['append', ' .= ']):
                res += self.finding(i.Origin, i.InFileLine,
                                    override_msg='Append to list should start with a blank')
            if not i.VarValue.endswith(' "') and any(x in ops for x in ['prepend', ' =. ']):
                res += self.finding(i.Origin, i.InFileLine,
                                    override_msg='Prepend to list should end with a blank')
        return res



import subprocess

import setuptools

_long_description = 'See https://github.com/priv-kweihmann/oelint-adv for documentation'
_long_description_content_type = 'text/plain'
try:
    _long_description = subprocess.check_output(
        ['pandoc', '--from', 'markdown', '--to', 'markdown', 'README.md']).decode('utf-8')
    _long_description_content_type = 'text/markdown'
except (subprocess.CalledProcessError, FileNotFoundError):
    pass

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name='oelint_adv',
    version='3.9.2',
    author='Konrad Weihmann',
    author_email='kweihmann@outlook.com',
    description='Advanced bitbake-recipe linter',
    long_description=_long_description,
    long_description_content_type=_long_description_content_type,
    url='https://github.com/priv-kweihmann/oelint-adv',
    packages=setuptools.find_packages(exclude=('tests',)),
    entry_points={
        'console_scripts': [
            'oelint-adv = oelint_adv.__main__:main',
        ],
    },
    install_requires=requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Quality Assurance',
    ],
)



from oelint_parser.cls_item import Function
from oelint_adv.cls_rule import Rule


class TaskInstallNoMkdir(Rule):
    def __init__(self):
        super().__init__(id='oelint.task.nomkdir',
                         severity='error',
                         message='\'mkdir\' shall not be used in do_install. Use \'install\'')

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER):
            if item.FuncName.startswith('do_install'):
                if item.FuncBody.find(' mkdir ') != -1:
                    res += self.finding(item.Origin, item.InFileLine)
        return res



import re

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Include


class VarMultiInclude(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.multiinclude',
                         severity='warning',
                         message='\'{INC}\' is included multiple times')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Include.CLASSIFIER)
        keys = []
        for i in items:
            keys += [x.strip() for x in re.split(r'\s|,', i.IncName) if x]
        for key in list(set(keys)):
            _i = [x for x in items if x.IncName.find(key) != -1]
            if len(_i) > 1:
                res += self.finding(_i[-1].Origin, _i[-1].InFileLine,
                                    self.Msg.replace('{INC}', key))
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import get_scr_components


class VarPnBpnUsage(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.pnbpnusage',
                         severity='error',
                         message='${BPN} should be used instead of ${PN}')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        needles = ['SRC_URI']
        for i in [x for x in items if x.VarName in needles]:
            for x in i.get_items():
                _comp = get_scr_components(x)
                if '${PN}' in _comp['src']:
                    res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import get_scr_components


class VarBugtrackerIsUrl(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.bugtrackerisurl',
                         severity='warning',
                         message='\'BUGTRACKER\' should be an URL')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='BUGTRACKER')
        for i in items:
            val = i.VarValueStripped
            try:
                result = get_scr_components(val)
                if not result['scheme'] or not result['src']:
                    raise Exception()
            except Exception:
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarInsaneSkip(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.insaneskip',
                         severity='error',
                         message='INSANE_SKIP should be avoided at any cost')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        for i in [x for x in items if x.VarName.startswith('INSANE_SKIP_') or x.VarName == 'INSANE_SKIP']:
            res += self.finding(i.Origin, i.InFileLine)
        return res
